import numpy as np
import pytest

from autolens import exc
from autolens.data.array import grids
from autolens.data.array import scaled_array as sca
from autolens.model.galaxy import galaxy as g, galaxy_data as gd
from autolens.model.profiles import light_profiles as lp
from autolens.model.profiles import mass_profiles as mp

from test.unit.mock.model import mock_galaxy


class TestGalaxyFitData(object):
    def test__image_noise_map_and_mask(self, gal_data_7x7, mask_7x7):

        galaxy_fit_data = gd.GalaxyFitData(
            galaxy_data=gal_data_7x7, mask=mask_7x7, use_intensities=True
        )

        assert galaxy_fit_data.pixel_scale == 1.0
        assert (galaxy_fit_data.unmasked_image == np.ones((7, 7))).all()
        assert (galaxy_fit_data.unmasked_noise_map == 2.0 * np.ones((7, 7))).all()

        assert (galaxy_fit_data.image_1d == np.ones(9)).all()
        assert (galaxy_fit_data.noise_map_1d == 2.0 * np.ones(9)).all()
        assert (galaxy_fit_data.mask_1d == np.full(fill_value=False, shape=(9))).all()

        assert (
            galaxy_fit_data.mask(return_in_2d=True)
            == np.array(
                [
                    [True, True, True, True, True, True, True],
                    [True, True, True, True, True, True, True],
                    [True, True, False, False, False, True, True],
                    [True, True, False, False, False, True, True],
                    [True, True, False, False, False, True, True],
                    [True, True, True, True, True, True, True],
                    [True, True, True, True, True, True, True],
                ]
            )
        ).all()

        assert (
            galaxy_fit_data.image(return_in_2d=True)
            == np.array(
                [
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 1.0, 1.0, 1.0, 0.0, 0.0],
                    [0.0, 0.0, 1.0, 1.0, 1.0, 0.0, 0.0],
                    [0.0, 0.0, 1.0, 1.0, 1.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                ]
            )
        ).all()

        assert (
            galaxy_fit_data.noise_map(return_in_2d=True)
            == np.array(
                [
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 2.0, 2.0, 2.0, 0.0, 0.0],
                    [0.0, 0.0, 2.0, 2.0, 2.0, 0.0, 0.0],
                    [0.0, 0.0, 2.0, 2.0, 2.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                ]
            )
        ).all()

    def test__grid_stack(
        self, gal_data_7x7, mask_7x7, grid_7x7, sub_grid_7x7, blurring_grid_7x7
    ):

        galaxy_fit_data = gd.GalaxyFitData(
            galaxy_data=gal_data_7x7, mask=mask_7x7, use_intensities=True
        )

        assert galaxy_fit_data.grid_stack.regular.shape == (9, 2)

        assert (galaxy_fit_data.grid_stack.regular == grid_7x7).all()
        assert (galaxy_fit_data.grid_stack.sub == sub_grid_7x7).all()

    def test__interp_pixel_scale(self, image_7x7, mask_7x7):

        noise_map = sca.ScaledSquarePixelArray(
            array=2.0 * np.ones((7, 7)), pixel_scale=3.0
        )
        gal_data_7x7 = gd.GalaxyData(
            image=image_7x7, noise_map=noise_map, pixel_scale=3.0
        )
        gal_data_7x7 = gd.GalaxyFitData(
            galaxy_data=gal_data_7x7,
            mask=mask_7x7,
            interp_pixel_scale=1.0,
            use_intensities=True,
        )

        grid_stack = grids.GridStack.grid_stack_from_mask_sub_grid_size_and_psf_shape(
            mask=mask_7x7, sub_grid_size=2, psf_shape=(3, 3)
        )
        new_grid_stack = grid_stack.new_grid_stack_with_interpolator_added_to_each_grid(
            interp_pixel_scale=1.0
        )

        assert (gal_data_7x7.grid_stack.regular == new_grid_stack.regular).all()
        assert (
            gal_data_7x7.grid_stack.regular.interpolator.vtx
            == new_grid_stack.regular.interpolator.vtx
        ).all()
        assert (
            gal_data_7x7.grid_stack.regular.interpolator.wts
            == new_grid_stack.regular.interpolator.wts
        ).all()

        assert (gal_data_7x7.grid_stack.sub == new_grid_stack.sub).all()
        assert (
            gal_data_7x7.grid_stack.sub.interpolator.vtx
            == new_grid_stack.sub.interpolator.vtx
        ).all()
        assert (
            gal_data_7x7.grid_stack.sub.interpolator.wts
            == new_grid_stack.sub.interpolator.wts
        ).all()

        assert (gal_data_7x7.grid_stack.blurring == new_grid_stack.blurring).all()
        assert (
            gal_data_7x7.grid_stack.blurring.interpolator.vtx
            == new_grid_stack.blurring.interpolator.vtx
        ).all()
        assert (
            gal_data_7x7.grid_stack.blurring.interpolator.wts
            == new_grid_stack.blurring.interpolator.wts
        ).all()

    def test__gal_data_7x7_intensities(self, gal_data_7x7, mask_7x7):

        galaxy_fit_data = gd.GalaxyFitData(
            galaxy_data=gal_data_7x7, mask=mask_7x7, use_intensities=True
        )

        assert galaxy_fit_data.pixel_scale == 1.0
        assert (galaxy_fit_data.unmasked_image == np.ones((7, 7))).all()
        assert (galaxy_fit_data.unmasked_noise_map == 2.0 * np.ones((7, 7))).all()

        assert (galaxy_fit_data.image_1d == np.ones(9)).all()
        assert (galaxy_fit_data.noise_map_1d == 2.0 * np.ones(9)).all()
        assert (galaxy_fit_data.mask_1d == np.full(fill_value=False, shape=(9))).all()

        assert (
            galaxy_fit_data.mask(return_in_2d=True)
            == np.array(
                [
                    [True, True, True, True, True, True, True],
                    [True, True, True, True, True, True, True],
                    [True, True, False, False, False, True, True],
                    [True, True, False, False, False, True, True],
                    [True, True, False, False, False, True, True],
                    [True, True, True, True, True, True, True],
                    [True, True, True, True, True, True, True],
                ]
            )
        ).all()

        assert (
            galaxy_fit_data.image(return_in_2d=True)
            == np.array(
                [
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 1.0, 1.0, 1.0, 0.0, 0.0],
                    [0.0, 0.0, 1.0, 1.0, 1.0, 0.0, 0.0],
                    [0.0, 0.0, 1.0, 1.0, 1.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                ]
            )
        ).all()

        assert (
            galaxy_fit_data.noise_map(return_in_2d=True)
            == np.array(
                [
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 2.0, 2.0, 2.0, 0.0, 0.0],
                    [0.0, 0.0, 2.0, 2.0, 2.0, 0.0, 0.0],
                    [0.0, 0.0, 2.0, 2.0, 2.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                ]
            )
        ).all()

        galaxy = mock_galaxy.MockGalaxy(value=1, shape=36)

        intensities = galaxy_fit_data.profile_quantity_from_galaxy_and_sub_grid(
            galaxies=[galaxy], sub_grid=galaxy_fit_data.grid_stack.sub
        )

        assert (intensities == np.ones(9)).all()

        galaxy = g.Galaxy(redshift=0.5, light=lp.SphericalSersic(intensity=1.0))

        intensities_gal = galaxy.intensities_from_grid(
            grid=galaxy_fit_data.grid_stack.sub, return_in_2d=False, return_binned=True
        )

        intensities_gd = galaxy_fit_data.profile_quantity_from_galaxy_and_sub_grid(
            galaxies=[galaxy], sub_grid=galaxy_fit_data.grid_stack.sub
        )

        assert (intensities_gal == intensities_gd).all()

    def test__gal_data_7x7_convergence(self, gal_data_7x7, mask_7x7):

        galaxy_fit_data = gd.GalaxyFitData(
            galaxy_data=gal_data_7x7,
            mask=mask_7x7,
            sub_grid_size=2,
            use_convergence=True,
        )

        assert galaxy_fit_data.pixel_scale == 1.0
        assert (galaxy_fit_data.unmasked_image == np.ones((7, 7))).all()
        assert (galaxy_fit_data.unmasked_noise_map == 2.0 * np.ones((7, 7))).all()

        assert (galaxy_fit_data.image_1d == np.ones(9)).all()
        assert (galaxy_fit_data.noise_map_1d == 2.0 * np.ones(9)).all()
        assert (galaxy_fit_data.mask_1d == np.full(fill_value=False, shape=(9))).all()

        assert (
            galaxy_fit_data.mask(return_in_2d=True)
            == np.array(
                [
                    [True, True, True, True, True, True, True],
                    [True, True, True, True, True, True, True],
                    [True, True, False, False, False, True, True],
                    [True, True, False, False, False, True, True],
                    [True, True, False, False, False, True, True],
                    [True, True, True, True, True, True, True],
                    [True, True, True, True, True, True, True],
                ]
            )
        ).all()

        assert (
            galaxy_fit_data.image(return_in_2d=True)
            == np.array(
                [
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 1.0, 1.0, 1.0, 0.0, 0.0],
                    [0.0, 0.0, 1.0, 1.0, 1.0, 0.0, 0.0],
                    [0.0, 0.0, 1.0, 1.0, 1.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                ]
            )
        ).all()

        assert (
            galaxy_fit_data.noise_map(return_in_2d=True)
            == np.array(
                [
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 2.0, 2.0, 2.0, 0.0, 0.0],
                    [0.0, 0.0, 2.0, 2.0, 2.0, 0.0, 0.0],
                    [0.0, 0.0, 2.0, 2.0, 2.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                ]
            )
        ).all()

        galaxy = mock_galaxy.MockGalaxy(value=1, shape=36)

        convergence = galaxy_fit_data.profile_quantity_from_galaxy_and_sub_grid(
            galaxies=[galaxy], sub_grid=galaxy_fit_data.grid_stack.sub
        )

        assert (convergence == np.ones(9)).all()

        galaxy = g.Galaxy(
            redshift=0.5, mass=mp.SphericalIsothermal(einstein_radius=1.0)
        )

        convergence_gal = galaxy.convergence_from_grid(
            grid=galaxy_fit_data.grid_stack.sub, return_in_2d=False, return_binned=True
        )

        convergence_gd = galaxy_fit_data.profile_quantity_from_galaxy_and_sub_grid(
            galaxies=[galaxy], sub_grid=galaxy_fit_data.grid_stack.sub
        )

        assert (convergence_gal == convergence_gd).all()

    def test__gal_data_7x7_potential(self, gal_data_7x7, mask_7x7):

        galaxy_fit_data = gd.GalaxyFitData(
            galaxy_data=gal_data_7x7, mask=mask_7x7, sub_grid_size=2, use_potential=True
        )

        assert galaxy_fit_data.pixel_scale == 1.0
        assert (galaxy_fit_data.unmasked_image == np.ones((7, 7))).all()
        assert (galaxy_fit_data.unmasked_noise_map == 2.0 * np.ones((7, 7))).all()

        assert (galaxy_fit_data.image_1d == np.ones(9)).all()
        assert (galaxy_fit_data.noise_map_1d == 2.0 * np.ones(9)).all()
        assert (galaxy_fit_data.mask_1d == np.full(fill_value=False, shape=(9))).all()

        assert (
            galaxy_fit_data.mask(return_in_2d=True)
            == np.array(
                [
                    [True, True, True, True, True, True, True],
                    [True, True, True, True, True, True, True],
                    [True, True, False, False, False, True, True],
                    [True, True, False, False, False, True, True],
                    [True, True, False, False, False, True, True],
                    [True, True, True, True, True, True, True],
                    [True, True, True, True, True, True, True],
                ]
            )
        ).all()

        assert (
            galaxy_fit_data.image(return_in_2d=True)
            == np.array(
                [
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 1.0, 1.0, 1.0, 0.0, 0.0],
                    [0.0, 0.0, 1.0, 1.0, 1.0, 0.0, 0.0],
                    [0.0, 0.0, 1.0, 1.0, 1.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                ]
            )
        ).all()

        assert (
            galaxy_fit_data.noise_map(return_in_2d=True)
            == np.array(
                [
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 2.0, 2.0, 2.0, 0.0, 0.0],
                    [0.0, 0.0, 2.0, 2.0, 2.0, 0.0, 0.0],
                    [0.0, 0.0, 2.0, 2.0, 2.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                ]
            )
        ).all()

        galaxy = mock_galaxy.MockGalaxy(value=1, shape=36)

        potential = galaxy_fit_data.profile_quantity_from_galaxy_and_sub_grid(
            galaxies=[galaxy], sub_grid=galaxy_fit_data.grid_stack.sub
        )

        assert (potential == np.ones(9)).all()

        galaxy = g.Galaxy(
            redshift=0.5, mass=mp.SphericalIsothermal(einstein_radius=1.0)
        )

        potential_gal = galaxy.potential_from_grid(
            grid=galaxy_fit_data.grid_stack.sub, return_binned=True
        )

        potential_gd = galaxy_fit_data.profile_quantity_from_galaxy_and_sub_grid(
            galaxies=[galaxy], sub_grid=galaxy_fit_data.grid_stack.sub
        )

        assert (potential_gal == potential_gd).all()

    def test__gal_data_7x7_deflections_y(self, gal_data_7x7, mask_7x7):

        galaxy_fit_data = gd.GalaxyFitData(
            galaxy_data=gal_data_7x7,
            mask=mask_7x7,
            sub_grid_size=2,
            use_deflections_y=True,
        )
        assert galaxy_fit_data.pixel_scale == 1.0
        assert (galaxy_fit_data.unmasked_image == np.ones((7, 7))).all()
        assert (galaxy_fit_data.unmasked_noise_map == 2.0 * np.ones((7, 7))).all()

        assert (galaxy_fit_data.image_1d == np.ones(9)).all()
        assert (galaxy_fit_data.noise_map_1d == 2.0 * np.ones(9)).all()
        assert (galaxy_fit_data.mask_1d == np.full(fill_value=False, shape=(9))).all()

        assert (
            galaxy_fit_data.mask(return_in_2d=True)
            == np.array(
                [
                    [True, True, True, True, True, True, True],
                    [True, True, True, True, True, True, True],
                    [True, True, False, False, False, True, True],
                    [True, True, False, False, False, True, True],
                    [True, True, False, False, False, True, True],
                    [True, True, True, True, True, True, True],
                    [True, True, True, True, True, True, True],
                ]
            )
        ).all()

        assert (
            galaxy_fit_data.image(return_in_2d=True)
            == np.array(
                [
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 1.0, 1.0, 1.0, 0.0, 0.0],
                    [0.0, 0.0, 1.0, 1.0, 1.0, 0.0, 0.0],
                    [0.0, 0.0, 1.0, 1.0, 1.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                ]
            )
        ).all()

        assert (
            galaxy_fit_data.noise_map(return_in_2d=True)
            == np.array(
                [
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 2.0, 2.0, 2.0, 0.0, 0.0],
                    [0.0, 0.0, 2.0, 2.0, 2.0, 0.0, 0.0],
                    [0.0, 0.0, 2.0, 2.0, 2.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                ]
            )
        ).all()

        galaxy = mock_galaxy.MockGalaxy(value=1, shape=36)

        deflections_y = galaxy_fit_data.profile_quantity_from_galaxy_and_sub_grid(
            galaxies=[galaxy], sub_grid=galaxy_fit_data.grid_stack.sub
        )

        assert (deflections_y == np.ones(9)).all()

        galaxy = g.Galaxy(
            redshift=0.5, mass=mp.SphericalIsothermal(einstein_radius=1.0)
        )

        deflections_gal = galaxy.deflections_from_grid(
            grid=galaxy_fit_data.grid_stack.sub
        )
        deflections_gal = np.asarray(
            [
                galaxy_fit_data.grid_stack.sub.array_1d_binned_from_sub_array_1d(
                    deflections_gal[:, 0]
                ),
                galaxy_fit_data.grid_stack.sub.array_1d_binned_from_sub_array_1d(
                    deflections_gal[:, 1]
                ),
            ]
        ).T

        deflections_gd = galaxy_fit_data.profile_quantity_from_galaxy_and_sub_grid(
            galaxies=[galaxy], sub_grid=galaxy_fit_data.grid_stack.sub
        )

        assert (deflections_gal[:, 0] == deflections_gd).all()

    def test__gal_data_7x7_deflections_x(self, gal_data_7x7, mask_7x7):

        galaxy_fit_data = gd.GalaxyFitData(
            galaxy_data=gal_data_7x7,
            mask=mask_7x7,
            sub_grid_size=2,
            use_deflections_x=True,
        )

        assert galaxy_fit_data.pixel_scale == 1.0
        assert (galaxy_fit_data.unmasked_image == np.ones((7, 7))).all()
        assert (galaxy_fit_data.unmasked_noise_map == 2.0 * np.ones((7, 7))).all()

        assert (galaxy_fit_data.image_1d == np.ones(9)).all()
        assert (galaxy_fit_data.noise_map_1d == 2.0 * np.ones(9)).all()
        assert (galaxy_fit_data.mask_1d == np.full(fill_value=False, shape=(9))).all()

        assert (
            galaxy_fit_data.mask(return_in_2d=True)
            == np.array(
                [
                    [True, True, True, True, True, True, True],
                    [True, True, True, True, True, True, True],
                    [True, True, False, False, False, True, True],
                    [True, True, False, False, False, True, True],
                    [True, True, False, False, False, True, True],
                    [True, True, True, True, True, True, True],
                    [True, True, True, True, True, True, True],
                ]
            )
        ).all()

        assert (
            galaxy_fit_data.image(return_in_2d=True)
            == np.array(
                [
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 1.0, 1.0, 1.0, 0.0, 0.0],
                    [0.0, 0.0, 1.0, 1.0, 1.0, 0.0, 0.0],
                    [0.0, 0.0, 1.0, 1.0, 1.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                ]
            )
        ).all()

        assert (
            galaxy_fit_data.noise_map(return_in_2d=True)
            == np.array(
                [
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 2.0, 2.0, 2.0, 0.0, 0.0],
                    [0.0, 0.0, 2.0, 2.0, 2.0, 0.0, 0.0],
                    [0.0, 0.0, 2.0, 2.0, 2.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                ]
            )
        ).all()

        galaxy = mock_galaxy.MockGalaxy(value=1, shape=36)

        deflections_x = galaxy_fit_data.profile_quantity_from_galaxy_and_sub_grid(
            galaxies=[galaxy], sub_grid=galaxy_fit_data.grid_stack.sub
        )

        assert (deflections_x == np.ones(9)).all()

        galaxy = g.Galaxy(
            redshift=0.5, mass=mp.SphericalIsothermal(einstein_radius=1.0)
        )

        deflections_gal = galaxy.deflections_from_grid(
            grid=galaxy_fit_data.grid_stack.sub
        )
        deflections_gal = np.asarray(
            [
                galaxy_fit_data.grid_stack.sub.array_1d_binned_from_sub_array_1d(
                    deflections_gal[:, 0]
                ),
                galaxy_fit_data.grid_stack.sub.array_1d_binned_from_sub_array_1d(
                    deflections_gal[:, 1]
                ),
            ]
        ).T

        deflections_gd = galaxy_fit_data.profile_quantity_from_galaxy_and_sub_grid(
            galaxies=[galaxy], sub_grid=galaxy_fit_data.grid_stack.sub
        )

        assert (deflections_gal[:, 1] == deflections_gd).all()

    def test__no_use_method__raises_exception(self, image_7x7, mask_7x7):

        gal_data_7x7 = gd.GalaxyData(
            image=image_7x7, noise_map=2.0 * np.ones((7, 7)), pixel_scale=3.0
        )

        with pytest.raises(exc.GalaxyException):
            gd.GalaxyFitData(galaxy_data=gal_data_7x7, mask=mask_7x7, sub_grid_size=2)

    def test__multiple_use_methods__raises_exception(self, image_7x7, mask_7x7):

        gal_data_7x7 = gd.GalaxyData(
            image=image_7x7, noise_map=2.0 * np.ones((7, 7)), pixel_scale=3.0
        )

        with pytest.raises(exc.GalaxyException):
            gd.GalaxyFitData(
                galaxy_data=gal_data_7x7,
                mask=mask_7x7,
                sub_grid_size=2,
                use_intensities=True,
                use_convergence=True,
            )

        with pytest.raises(exc.GalaxyException):
            gd.GalaxyFitData(
                galaxy_data=gal_data_7x7,
                mask=mask_7x7,
                sub_grid_size=2,
                use_intensities=True,
                use_potential=True,
            )

        with pytest.raises(exc.GalaxyException):
            gd.GalaxyFitData(
                galaxy_data=gal_data_7x7,
                mask=mask_7x7,
                sub_grid_size=2,
                use_intensities=True,
                use_deflections_y=True,
            )

        with pytest.raises(exc.GalaxyException):
            gd.GalaxyFitData(
                galaxy_data=gal_data_7x7,
                mask=mask_7x7,
                sub_grid_size=2,
                use_intensities=True,
                use_convergence=True,
                use_potential=True,
            )

        with pytest.raises(exc.GalaxyException):
            gd.GalaxyFitData(
                galaxy_data=gal_data_7x7,
                mask=mask_7x7,
                sub_grid_size=2,
                use_intensities=True,
                use_convergence=True,
                use_potential=True,
                use_deflections_x=True,
            )



import autofit as af

from autolens.data.array.grids import reshape_returned_regular_array


class GalaxyFit(af.DataFit1D):
    def __init__(self, galaxy_data, model_galaxies):
        """Class which fits a set of galaxy-datas to a model galaxy, using either the galaxy's intensities, \
        surface-density or potential.

        Parameters
        ----------
        galaxy_data : GalaxyData
            The galaxy-datas object being fitted.
        model_galaxies : galaxy.Galaxy
            The model galaxy used to fit the galaxy-datas.
        """

        self.galaxy_data = galaxy_data
        self.mask_2d = galaxy_data.mask_2d
        self.model_galaxies = model_galaxies
        self.map_to_scaled_array = galaxy_data.grid_stack.scaled_array_2d_from_array_1d

        model_data_1d = galaxy_data.profile_quantity_from_galaxy_and_sub_grid(
            galaxies=model_galaxies, sub_grid=galaxy_data.grid_stack.sub
        )

        super(GalaxyFit, self).__init__(
            data_1d=galaxy_data.image_1d,
            noise_map_1d=galaxy_data.noise_map_1d,
            mask_1d=galaxy_data.mask_1d,
            model_data_1d=model_data_1d,
        )

    @property
    def grid_stack(self):
        return self.galaxy_data.grid_stack

    @property
    def image_1d(self):
        return self.data_1d

    @property
    def model_image_1d(self):
        return self.model_data_1d

    def mask(self, return_in_2d=True):
        if return_in_2d:
            return self.mask_2d
        else:
            return self.mask_1d

    @reshape_returned_regular_array
    def image(self):
        return self.image_1d

    @reshape_returned_regular_array
    def noise_map(self):
        return self.noise_map_1d

    @reshape_returned_regular_array
    def signal_to_noise_map(self):
        return self.signal_to_noise_map_1d

    @reshape_returned_regular_array
    def model_image(self):
        return self.model_image_1d

    @reshape_returned_regular_array
    def residual_map(self):
        return self.residual_map_1d

    @reshape_returned_regular_array
    def normalized_residual_map(self):
        return self.normalized_residual_map_1d

    @reshape_returned_regular_array
    def chi_squared_map(self):
        return self.chi_squared_map_1d

    @property
    def figure_of_merit(self):
        return self.likelihood



import os

import numpy as np

from autolens.data.array.util import mapping_util, mask_util
from autolens.data.array import mask

test_data_dir = "{}/../test_files/array/".format(
    os.path.dirname(os.path.realpath(__file__))
)


class TestMap1dIndexesTo2dIndex(object):
    def test__9_1d_indexes_from_0_to_8__map_to_shape_3x3(self):

        indexes_1d = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8])

        indexes_2d = mapping_util.map_1d_indexes_to_2d_indexes_for_shape(
            indexes_1d=indexes_1d, shape=(3, 3)
        )

        assert (
            indexes_2d
            == np.array(
                [[0, 0], [0, 1], [0, 2], [1, 0], [1, 1], [1, 2], [2, 0], [2, 1], [2, 2]]
            )
        ).all()

    def test__6_1d_indexes_from_0_to_5__map_to_shape_2x3(self):

        indexes_1d = np.array([0, 1, 2, 3, 4, 5])

        indexes_2d = mapping_util.map_1d_indexes_to_2d_indexes_for_shape(
            indexes_1d=indexes_1d, shape=(2, 3)
        )

        assert (
            indexes_2d == np.array([[0, 0], [0, 1], [0, 2], [1, 0], [1, 1], [1, 2]])
        ).all()

    def test__6_1d_indexes_from_0_to_5__map_to_shape_3x2(self):

        indexes_1d = np.array([0, 1, 2, 3, 4, 5])

        indexes_2d = mapping_util.map_1d_indexes_to_2d_indexes_for_shape(
            indexes_1d=indexes_1d, shape=(3, 2)
        )

        assert (
            indexes_2d == np.array([[0, 0], [0, 1], [1, 0], [1, 1], [2, 0], [2, 1]])
        ).all()

    def test__9_1d_indexes_from_0_to_8_different_order__map_to_shape_3x3(self):

        indexes_1d = np.array([1, 4, 7, 8, 0, 2, 3, 5, 6])

        indexes_2d = mapping_util.map_1d_indexes_to_2d_indexes_for_shape(
            indexes_1d=indexes_1d, shape=(3, 3)
        )

        assert (
            indexes_2d
            == np.array(
                [[0, 1], [1, 1], [2, 1], [2, 2], [0, 0], [0, 2], [1, 0], [1, 2], [2, 0]]
            )
        ).all()


class TestMap2dIndexesTo1dIndex(object):
    def test__9_2d_indexes_from_0_0_to_2_2__map_to_shape_3x3(self):

        indexes_2d = np.array(
            [[0, 0], [0, 1], [0, 2], [1, 0], [1, 1], [1, 2], [2, 0], [2, 1], [2, 2]]
        )

        indexes_1d = mapping_util.map_2d_indexes_to_1d_indexes_for_shape(
            indexes_2d=indexes_2d, shape=(3, 3)
        )

        assert (indexes_1d == np.array([0, 1, 2, 3, 4, 5, 6, 7, 8])).all()

    def test__6_1d_indexes_from_0_0_to_1_2__map_to_shape_2x3(self):

        indexes_2d = np.array([[0, 0], [0, 1], [0, 2], [1, 0], [1, 1], [1, 2]])

        indexes_1d = mapping_util.map_2d_indexes_to_1d_indexes_for_shape(
            indexes_2d=indexes_2d, shape=(2, 3)
        )

        assert (indexes_1d == np.array([0, 1, 2, 3, 4, 5])).all()

    def test__6_1d_indexes_from_0_0_to_2_1__map_to_shape_3x2(self):

        indexes_2d = np.array([[0, 0], [0, 1], [1, 0], [1, 1], [2, 0], [2, 1]])

        indexes_1d = mapping_util.map_2d_indexes_to_1d_indexes_for_shape(
            indexes_2d=indexes_2d, shape=(3, 2)
        )

        assert (indexes_1d == np.array([0, 1, 2, 3, 4, 5])).all()

    def test__9_1d_indexes_from_0_0_to_2_2_different_order__map_to_shape_3x3(self):

        indexes_2d = np.array(
            [[0, 1], [1, 1], [2, 1], [2, 2], [0, 0], [0, 2], [1, 0], [1, 2], [2, 0]]
        )

        indexes_1d = mapping_util.map_2d_indexes_to_1d_indexes_for_shape(
            indexes_2d=indexes_2d, shape=(3, 3)
        )

        assert (indexes_1d == np.array([1, 4, 7, 8, 0, 2, 3, 5, 6])).all()


class TestSubToImage(object):
    def test__3x3_mask_with_1_pixel__2x2_sub_grid__correct_sub_to_image(self):
        mask = np.array([[True, True, True], [True, False, True], [True, True, True]])

        sub_to_image = mapping_util.sub_to_regular_from_mask(mask, sub_grid_size=2)

        assert (sub_to_image == np.array([0, 0, 0, 0])).all()

    def test__3x3_mask_with_row_of_pixels_pixel__2x2_sub_grid__correct_sub_to_image(
        self
    ):
        mask = np.array([[True, True, True], [False, False, False], [True, True, True]])

        sub_to_image = mapping_util.sub_to_regular_from_mask(mask, sub_grid_size=2)

        assert (sub_to_image == np.array([0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2])).all()

    def test__3x3_mask_with_row_of_pixels_pixel__3x3_sub_grid__correct_sub_to_image(
        self
    ):
        mask = np.array([[True, True, True], [False, False, False], [True, True, True]])

        sub_to_image = mapping_util.sub_to_regular_from_mask(mask, sub_grid_size=3)

        assert (
            sub_to_image
            == np.array(
                [
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    1,
                    2,
                    2,
                    2,
                    2,
                    2,
                    2,
                    2,
                    2,
                    2,
                ]
            )
        ).all()


class TestSubArray1DFromSubArray2d(object):
    def test__map_simple_data__sub_grid_size_1(self):

        array_2d = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])

        mask = np.array([[True, True, True], [True, False, True], [True, True, True]])

        array_1d = mapping_util.sub_array_1d_from_sub_array_2d_mask_and_sub_grid_size(
            mask=mask, sub_array_2d=array_2d, sub_grid_size=1
        )

        assert (array_1d == np.array([5])).all()

        array_2d = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])

        mask = np.array(
            [[True, False, True], [False, False, False], [True, False, True]]
        )

        array_1d = mapping_util.sub_array_1d_from_sub_array_2d_mask_and_sub_grid_size(
            mask=mask, sub_array_2d=array_2d, sub_grid_size=1
        )

        assert (array_1d == np.array([2, 4, 5, 6, 8])).all()

        array_2d = np.array([[1, 2, 3, 4], [5, 6, 7, 8], [9, 10, 11, 12]])

        mask = np.array(
            [
                [True, False, True, True],
                [False, False, False, True],
                [True, False, True, False],
            ]
        )

        array_1d = mapping_util.sub_array_1d_from_sub_array_2d_mask_and_sub_grid_size(
            mask=mask, sub_array_2d=array_2d, sub_grid_size=1
        )

        assert (array_1d == np.array([2, 5, 6, 7, 10, 12])).all()

        array_2d = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9], [10, 11, 12]])

        mask = np.array(
            [
                [True, False, True],
                [False, False, False],
                [True, False, True],
                [True, True, True],
            ]
        )

        array_1d = mapping_util.sub_array_1d_from_sub_array_2d_mask_and_sub_grid_size(
            mask=mask, sub_array_2d=array_2d, sub_grid_size=1
        )

        assert (array_1d == np.array([2, 4, 5, 6, 8])).all()

    def test__map_simple_data__sub_grid_size_2(self):

        sub_array_2d = np.array(
            [
                [1, 2, 3, 4, 5, 6],
                [7, 8, 9, 10, 11, 12],
                [13, 14, 15, 16, 17, 18],
                [1, 2, 3, 4, 5, 6],
                [7, 8, 9, 10, 11, 12],
                [13, 14, 15, 16, 17, 18],
            ]
        )

        mask = np.array([[True, True, True], [True, False, True], [True, True, True]])

        sub_array_1d = mapping_util.sub_array_1d_from_sub_array_2d_mask_and_sub_grid_size(
            sub_array_2d=sub_array_2d, mask=mask, sub_grid_size=2
        )

        assert (sub_array_1d == np.array([15, 16, 3, 4])).all()

        mask = np.array([[True, False, True], [True, False, True], [True, True, False]])

        sub_array_1d = mapping_util.sub_array_1d_from_sub_array_2d_mask_and_sub_grid_size(
            sub_array_2d=sub_array_2d, mask=mask, sub_grid_size=2
        )

        assert (
            sub_array_1d == np.array([3, 4, 9, 10, 15, 16, 3, 4, 11, 12, 17, 18])
        ).all()

        sub_array_2d = np.array(
            [
                [1, 2, 3, 4, 5, 6, 7, 7, 7],
                [7, 8, 9, 10, 11, 12, 7, 7, 7],
                [13, 14, 15, 16, 17, 18, 7, 7, 7],
                [1, 2, 3, 4, 5, 6, 7, 7, 7],
                [7, 8, 9, 10, 11, 12, 7, 7, 7],
                [13, 14, 15, 16, 17, 18, 7, 7, 7],
            ]
        )

        mask = np.array(
            [
                [True, False, True, True],
                [False, False, False, True],
                [True, False, True, False],
            ]
        )

        sub_array_1d = mapping_util.sub_array_1d_from_sub_array_2d_mask_and_sub_grid_size(
            sub_array_2d=sub_array_2d, mask=mask, sub_grid_size=2
        )

        assert (
            sub_array_1d
            == np.array(
                [
                    3,
                    4,
                    9,
                    10,
                    13,
                    14,
                    1,
                    2,
                    15,
                    16,
                    3,
                    4,
                    17,
                    18,
                    5,
                    6,
                    9,
                    10,
                    15,
                    16,
                    7,
                    7,
                    7,
                    7,
                ]
            )
        ).all()

        sub_array_2d = np.array(
            [
                [1, 2, 3, 4, 5, 6],
                [7, 8, 9, 10, 11, 12],
                [13, 14, 15, 16, 17, 18],
                [1, 2, 3, 4, 5, 6],
                [7, 8, 9, 10, 11, 12],
                [13, 14, 15, 16, 17, 18],
                [7, 7, 7, 7, 7, 7],
                [7, 7, 7, 7, 7, 7],
            ]
        )

        mask = np.array(
            [
                [True, False, True],
                [False, False, False],
                [True, False, True],
                [True, True, True],
            ]
        )

        sub_array_1d = mapping_util.sub_array_1d_from_sub_array_2d_mask_and_sub_grid_size(
            sub_array_2d=sub_array_2d, mask=mask, sub_grid_size=2
        )

        assert (
            sub_array_1d
            == np.array(
                [3, 4, 9, 10, 13, 14, 1, 2, 15, 16, 3, 4, 17, 18, 5, 6, 9, 10, 15, 16]
            )
        ).all()

    def test__setup_2x2_data__sub_grid_size_3(self):

        sub_array_2d = np.array(
            [
                [1, 2, 3, 4, 5, 6],
                [7, 8, 9, 10, 11, 12],
                [13, 14, 15, 16, 17, 18],
                [1, 2, 3, 4, 5, 6],
                [7, 8, 9, 10, 11, 12],
                [13, 14, 15, 16, 17, 18],
            ]
        )

        mask = np.array([[False, True], [True, False]])

        sub_array_1d = mapping_util.sub_array_1d_from_sub_array_2d_mask_and_sub_grid_size(
            sub_array_2d=sub_array_2d, mask=mask, sub_grid_size=3
        )

        assert (
            sub_array_1d
            == np.array([1, 2, 3, 7, 8, 9, 13, 14, 15, 4, 5, 6, 10, 11, 12, 16, 17, 18])
        ).all()


class TestSubGrid1DFromSubGrid2D(object):
    def test__map_simple_grids__sub_grid_1(self):

        grid_2d = np.array(
            [
                [[1, 1], [2, 2], [3, 3]],
                [[4, 4], [5, 5], [6, 6]],
                [[7, 7], [8, 8], [9, 9]],
            ]
        )

        mask = np.array([[True, True, True], [True, False, True], [True, True, True]])

        grid_1d = mapping_util.sub_grid_1d_from_sub_grid_2d_mask_and_sub_grid_size(
            sub_grid_2d=grid_2d, mask=mask, sub_grid_size=1
        )

        assert (grid_1d == np.array([[5, 5]])).all()

        grid_2d = np.array(
            [
                [[1, 1], [2, 2], [3, 3]],
                [[4, 4], [5, 5], [6, 6]],
                [[7, 7], [8, 8], [9, 9]],
            ]
        )

        mask = np.array(
            [[True, False, True], [False, False, False], [True, False, True]]
        )

        grid_1d = mapping_util.sub_grid_1d_from_sub_grid_2d_mask_and_sub_grid_size(
            sub_grid_2d=grid_2d, mask=mask, sub_grid_size=1
        )

        assert (grid_1d == np.array([[2, 2], [4, 4], [5, 5], [6, 6], [8, 8]])).all()

        grid_2d = np.array(
            [
                [[1, 1], [2, 2], [3, 3], [4, 4]],
                [[5, 5], [6, 6], [7, 7], [8, 8]],
                [[9, 9], [10, 10], [11, 11], [12, 12]],
            ]
        )

        mask = np.array(
            [
                [True, False, True, True],
                [False, False, False, True],
                [True, False, True, False],
            ]
        )

        grid_1d = mapping_util.sub_grid_1d_from_sub_grid_2d_mask_and_sub_grid_size(
            sub_grid_2d=grid_2d, mask=mask, sub_grid_size=1
        )

        assert (
            grid_1d == np.array([[2, 2], [5, 5], [6, 6], [7, 7], [10, 10], [12, 12]])
        ).all()

        grid_2d = np.array(
            [
                [[1, 1], [2, 2], [3, 3]],
                [[4, 4], [5, 5], [6, 6]],
                [[7, 7], [8, 8], [9, 9]],
                [[10, 10], [11, 11], [12, 12]],
            ]
        )

        mask = np.array(
            [
                [True, False, True],
                [False, False, False],
                [True, False, True],
                [True, True, True],
            ]
        )

        grid_1d = mapping_util.sub_grid_1d_from_sub_grid_2d_mask_and_sub_grid_size(
            sub_grid_2d=grid_2d, mask=mask, sub_grid_size=1
        )

        assert (grid_1d == np.array([[2, 2], [4, 4], [5, 5], [6, 6], [8, 8]])).all()

    def test__map_simple_grids__sub_grid_2(self):

        sub_grid_2d = np.array(
            [
                [[1, 1], [2, 2], [3, 3], [4, 4], [5, 5], [6, 6]],
                [[7, 7], [8, 8], [9, 9], [10, 10], [11, 11], [12, 12]],
                [[13, 13], [14, 14], [15, 15], [16, 16], [17, 17], [18, 18]],
                [[1, 1], [2, 2], [3, 3], [4, 4], [5, 5], [6, 6]],
                [[7, 7], [8, 8], [9, 9], [10, 10], [11, 11], [12, 12]],
                [[13, 13], [14, 14], [15, 15], [16, 16], [17, 17], [18, 18]],
            ]
        )

        mask = np.array([[True, False, True], [True, False, True], [True, True, False]])

        sub_array_1d = mapping_util.sub_grid_1d_from_sub_grid_2d_mask_and_sub_grid_size(
            sub_grid_2d=sub_grid_2d, mask=mask, sub_grid_size=2
        )

        assert (
            sub_array_1d
            == np.array(
                [
                    [3, 3],
                    [4, 4],
                    [9, 9],
                    [10, 10],
                    [15, 15],
                    [16, 16],
                    [3, 3],
                    [4, 4],
                    [11, 11],
                    [12, 12],
                    [17, 17],
                    [18, 18],
                ]
            )
        ).all()


class TestSubArray2dFromSubArray1d(object):
    def test__simple_2d_array__is_masked_and_mapped__sub_grid_size_1(self):

        array_1d = np.array([1.0, 2.0, 3.0, 4.0])

        mask = np.full(fill_value=False, shape=(2, 2))

        array_2d = mapping_util.sub_array_2d_from_sub_array_1d_mask_and_sub_grid_size(
            sub_array_1d=array_1d, mask=mask, sub_grid_size=1
        )

        assert (array_2d == np.array([[1.0, 2.0], [3.0, 4.0]])).all()

        array_1d = np.array([1.0, 2.0, 3.0])

        mask = np.array([[False, False], [False, True]])

        array_2d = mapping_util.sub_array_2d_from_sub_array_1d_mask_and_sub_grid_size(
            sub_array_1d=array_1d, mask=mask, sub_grid_size=1
        )

        assert (array_2d == np.array([[1.0, 2.0], [3.0, 0.0]])).all()

        array_1d = np.array([1.0, 2.0, 3.0, -1.0, -2.0, -3.0])

        mask = np.array(
            [
                [False, False, True, True],
                [False, True, True, True],
                [False, False, True, False],
            ]
        )

        array_2d = mapping_util.sub_array_2d_from_sub_array_1d_mask_and_sub_grid_size(
            sub_array_1d=array_1d, mask=mask, sub_grid_size=1
        )

        assert (
            array_2d
            == np.array(
                [[1.0, 2.0, 0.0, 0.0], [3.0, 0.0, 0.0, 0.0], [-1.0, -2.0, 0.0, -3.0]]
            )
        ).all()

    def test__simple_2d_array__is_masked_and_mapped__sub_grid_size_2(self):

        array_1d = np.array(
            [1.0, 1.0, 1.0, 1.0, 2.0, 2.0, 2.0, 2.0, 3.0, 3.0, 3.0, 4.0]
        )

        mask = np.array([[False, False], [False, True]])

        array_2d = mapping_util.sub_array_2d_from_sub_array_1d_mask_and_sub_grid_size(
            sub_array_1d=array_1d, mask=mask, sub_grid_size=2
        )

        assert (
            array_2d
            == np.array(
                [
                    [1.0, 1.0, 2.0, 2.0],
                    [1.0, 1.0, 2.0, 2.0],
                    [3.0, 3.0, 0.0, 0.0],
                    [3.0, 4.0, 0.0, 0.0],
                ]
            )
        ).all()


class TestSubGrid2DFromSubGrid1d(object):
    def test__simple_2d_array__is_masked_and_mapped__sub_grid_size_1(self):

        grid_1d = np.array([[1.0, 1.0], [2.0, 2.0], [3.0, 3.0], [4.0, 4.0]])

        mask = np.full(fill_value=False, shape=(2, 2))

        grid_2d = mapping_util.sub_grid_2d_from_sub_grid_1d_mask_and_sub_grid_size(
            sub_grid_1d=grid_1d, mask=mask, sub_grid_size=1
        )

        assert (
            grid_2d == np.array([[[1.0, 1.0], [2.0, 2.0]], [[3.0, 3.0], [4.0, 4.0]]])
        ).all()

        grid_1d = np.array([[1.0, 1.0], [2.0, 2.0], [3.0, 3.0]])

        mask = np.array([[False, False], [False, True]])

        grid_2d = mapping_util.sub_grid_2d_from_sub_grid_1d_mask_and_sub_grid_size(
            sub_grid_1d=grid_1d, mask=mask, sub_grid_size=1
        )

        assert (
            grid_2d == np.array([[[1.0, 1.0], [2.0, 2.0]], [[3.0, 3.0], [0.0, 0.0]]])
        ).all()

        grid_1d = np.array(
            [
                [1.0, 1.0],
                [2.0, 2.0],
                [3.0, 3.0],
                [-1.0, -1.0],
                [-2.0, -2.0],
                [-3.0, -3.0],
            ]
        )

        mask = np.array(
            [
                [False, False, True, True],
                [False, True, True, True],
                [False, False, True, False],
            ]
        )

        grid_2d = mapping_util.sub_grid_2d_from_sub_grid_1d_mask_and_sub_grid_size(
            sub_grid_1d=grid_1d, mask=mask, sub_grid_size=1
        )

        assert (
            grid_2d
            == np.array(
                [
                    [[1.0, 1.0], [2.0, 2.0], [0.0, 0.0], [0.0, 0.0]],
                    [[3.0, 3.0], [0.0, 0.0], [0.0, 0.0], [0.0, 0.0]],
                    [[-1.0, -1.0], [-2.0, -2.0], [0.0, 0.0], [-3.0, -3.0]],
                ]
            )
        ).all()

    def test__simple_2d_grid__is_masked_and_mapped__sub_grid_size_2(self):

        grid_1d = np.array(
            [
                [1.0, 1.0],
                [1.0, 1.0],
                [1.0, 1.0],
                [1.0, 1.0],
                [2.0, 2.0],
                [2.0, 2.0],
                [2.0, 2.0],
                [2.0, 2.0],
                [3.0, 3.0],
                [3.0, 3.0],
                [3.0, 3.0],
                [4.0, 4.0],
            ]
        )

        mask = np.array([[False, False], [False, True]])

        grid_2d = mapping_util.sub_grid_2d_from_sub_grid_1d_mask_and_sub_grid_size(
            sub_grid_1d=grid_1d, mask=mask, sub_grid_size=2
        )

        assert (
            grid_2d
            == np.array(
                [
                    [[1.0, 1.0], [1.0, 1.0], [2.0, 2.0], [2.0, 2.0]],
                    [[1.0, 1.0], [1.0, 1.0], [2.0, 2.0], [2.0, 2.0]],
                    [[3.0, 3.0], [3.0, 3.0], [0.0, 0.0], [0.0, 0.0]],
                    [[3.0, 3.0], [4.0, 4.0], [0.0, 0.0], [0.0, 0.0]],
                ]
            )
        ).all()


class TestSparseToUnmaskedSparse:
    def test__mask_full_false__image_mask_and_pixel_centres_fully_overlap__each_sparse_maps_to_unmaked_sparse(
        self
    ):

        ma = mask.Mask(
            array=np.array(
                [[False, False, False], [False, False, False], [False, False, False]]
            ),
            pixel_scale=1.0,
        )

        unmasked_sparse_grid_pixel_centres = np.array(
            [[0, 0], [0, 1], [0, 2], [1, 0], [1, 1], [1, 2], [2, 0], [2, 1], [2, 2]]
        )

        total_masked_pixels = mask_util.total_sparse_pixels_from_mask(
            mask=ma,
            unmasked_sparse_grid_pixel_centres=unmasked_sparse_grid_pixel_centres,
        )

        sparse_to_unmasked_sparse = mapping_util.sparse_to_unmasked_sparse_from_mask_and_pixel_centres(
            total_sparse_pixels=total_masked_pixels,
            mask=ma,
            unmasked_sparse_grid_pixel_centres=unmasked_sparse_grid_pixel_centres,
        )

        assert (
            sparse_to_unmasked_sparse == np.array([0, 1, 2, 3, 4, 5, 6, 7, 8])
        ).all()

    def test__same_as_above__but_remove_some_centre_pixels_and_change_order__order_does_not_change_mapping(
        self
    ):

        ma = mask.Mask(
            array=np.array(
                [[False, False, False], [False, False, False], [False, False, False]]
            ),
            pixel_scale=1.0,
        )

        unmasked_sparse_grid_pixel_centres = np.array(
            [[0, 0], [0, 1], [2, 2], [1, 1], [0, 2], [2, 0], [0, 2]]
        )

        total_masked_pixels = mask_util.total_sparse_pixels_from_mask(
            mask=ma,
            unmasked_sparse_grid_pixel_centres=unmasked_sparse_grid_pixel_centres,
        )

        sparse_to_unmasked_sparse = mapping_util.sparse_to_unmasked_sparse_from_mask_and_pixel_centres(
            total_sparse_pixels=total_masked_pixels,
            mask=ma,
            unmasked_sparse_grid_pixel_centres=unmasked_sparse_grid_pixel_centres,
        )

        assert (sparse_to_unmasked_sparse == np.array([0, 1, 2, 3, 4, 5, 6])).all()

    def test__mask_is_cross__some_pix_pixels_are_masked__omitted_from_mapping(self):

        ma = mask.Mask(
            array=np.array(
                [[True, False, True], [False, False, False], [True, False, True]]
            ),
            pixel_scale=1.0,
        )

        unmasked_sparse_grid_pixel_centres = np.array(
            [[0, 0], [0, 1], [0, 2], [1, 0], [1, 1], [1, 2], [2, 0], [2, 1], [2, 2]]
        )

        total_masked_pixels = mask_util.total_sparse_pixels_from_mask(
            mask=ma,
            unmasked_sparse_grid_pixel_centres=unmasked_sparse_grid_pixel_centres,
        )

        sparse_to_unmasked_sparse = mapping_util.sparse_to_unmasked_sparse_from_mask_and_pixel_centres(
            total_sparse_pixels=total_masked_pixels,
            mask=ma,
            unmasked_sparse_grid_pixel_centres=unmasked_sparse_grid_pixel_centres,
        )

        assert (sparse_to_unmasked_sparse == np.array([1, 3, 4, 5, 7])).all()

    def test__same_as_above__different_mask_and_centres(self):

        ma = mask.Mask(
            array=np.array(
                [[False, False, True], [False, False, False], [True, False, False]]
            ),
            pixel_scale=1.0,
        )

        unmasked_sparse_grid_pixel_centres = np.array(
            [[0, 0], [0, 1], [0, 2], [0, 2], [0, 2], [1, 1]]
        )

        total_masked_pixels = mask_util.total_sparse_pixels_from_mask(
            mask=ma,
            unmasked_sparse_grid_pixel_centres=unmasked_sparse_grid_pixel_centres,
        )

        sparse_to_unmasked_sparse = mapping_util.sparse_to_unmasked_sparse_from_mask_and_pixel_centres(
            total_sparse_pixels=total_masked_pixels,
            mask=ma,
            unmasked_sparse_grid_pixel_centres=unmasked_sparse_grid_pixel_centres,
        )

        assert (sparse_to_unmasked_sparse == np.array([0, 1, 5])).all()

    def test__same_as_above__but_3x4_mask(self):

        ma = mask.Mask(
            array=np.array(
                [
                    [True, True, False, True],
                    [False, False, False, False],
                    [True, True, False, True],
                ]
            ),
            pixel_scale=1.0,
        )

        unmasked_sparse_grid_pixel_centres = np.array(
            [[0, 0], [0, 1], [0, 2], [0, 2], [0, 2], [1, 1], [2, 3], [2, 2]]
        )

        total_masked_pixels = mask_util.total_sparse_pixels_from_mask(
            mask=ma,
            unmasked_sparse_grid_pixel_centres=unmasked_sparse_grid_pixel_centres,
        )

        sparse_to_unmasked_sparse = mapping_util.sparse_to_unmasked_sparse_from_mask_and_pixel_centres(
            total_sparse_pixels=total_masked_pixels,
            mask=ma,
            unmasked_sparse_grid_pixel_centres=unmasked_sparse_grid_pixel_centres,
        )

        assert (sparse_to_unmasked_sparse == np.array([2, 3, 4, 5, 7])).all()

    def test__same_as_above__but_4x3_mask(self):

        ma = mask.Mask(
            array=np.array(
                [
                    [True, False, True],
                    [True, False, True],
                    [False, False, False],
                    [True, False, True],
                ]
            ),
            pixel_scale=1.0,
        )

        unmasked_sparse_grid_pixel_centres = np.array(
            [[0, 0], [0, 1], [0, 2], [0, 2], [0, 2], [1, 1], [2, 2], [3, 1]]
        )

        total_masked_pixels = mask_util.total_sparse_pixels_from_mask(
            mask=ma,
            unmasked_sparse_grid_pixel_centres=unmasked_sparse_grid_pixel_centres,
        )

        sparse_to_unmasked_sparse = mapping_util.sparse_to_unmasked_sparse_from_mask_and_pixel_centres(
            total_sparse_pixels=total_masked_pixels,
            mask=ma,
            unmasked_sparse_grid_pixel_centres=unmasked_sparse_grid_pixel_centres,
        )

        assert (sparse_to_unmasked_sparse == np.array([1, 5, 6, 7])).all()


class TestUnmaskedSparseToSparse:
    def test__mask_full_false__image_mask_and_pixel_centres_fully_overlap__each_pix_maps_to_unmaked_pix(
        self
    ):

        ma = mask.Mask(
            array=np.array(
                [[False, False, False], [False, False, False], [False, False, False]]
            ),
            pixel_scale=1.0,
        )

        unmasked_sparse_grid_pixel_centres = np.array(
            [[0, 0], [0, 1], [0, 2], [1, 0], [1, 1], [1, 2], [2, 0], [2, 1], [2, 2]]
        )

        unmasked_sparse_to_sparse = mapping_util.unmasked_sparse_to_sparse_from_mask_and_pixel_centres(
            mask=ma,
            unmasked_sparse_grid_pixel_centres=unmasked_sparse_grid_pixel_centres,
            total_sparse_pixels=9,
        )

        assert (
            unmasked_sparse_to_sparse == np.array([0, 1, 2, 3, 4, 5, 6, 7, 8])
        ).all()

    def test__same_as_above__but_remove_some_centre_pixels_and_change_order__order_does_not_change_mapping(
        self
    ):

        ma = mask.Mask(
            array=np.array(
                [[False, False, False], [False, False, False], [False, False, False]]
            ),
            pixel_scale=1.0,
        )

        unmasked_sparse_grid_pixel_centres = np.array(
            [[0, 0], [0, 1], [2, 2], [1, 1], [0, 2], [2, 0], [0, 2]]
        )

        unmasked_sparse_to_sparse = mapping_util.unmasked_sparse_to_sparse_from_mask_and_pixel_centres(
            mask=ma,
            unmasked_sparse_grid_pixel_centres=unmasked_sparse_grid_pixel_centres,
            total_sparse_pixels=9,
        )

        assert (unmasked_sparse_to_sparse == np.array([0, 1, 2, 3, 4, 5, 6])).all()

    def test__mask_is_cross__some_pix_pixels_are_masked__omitted_from_mapping(self):

        ma = mask.Mask(
            array=np.array(
                [[True, False, True], [False, False, False], [True, False, True]]
            ),
            pixel_scale=1.0,
        )

        unmasked_sparse_grid_pixel_centres = np.array(
            [[0, 0], [0, 1], [0, 2], [1, 0], [1, 1], [1, 2], [2, 0], [2, 1], [2, 2]]
        )

        unmasked_sparse_to_sparse = mapping_util.unmasked_sparse_to_sparse_from_mask_and_pixel_centres(
            mask=ma,
            unmasked_sparse_grid_pixel_centres=unmasked_sparse_grid_pixel_centres,
            total_sparse_pixels=5,
        )

        assert (
            unmasked_sparse_to_sparse == np.array([0, 0, 1, 1, 2, 3, 4, 4, 4])
        ).all()

    def test__same_as_above__different_mask_and_centres(self):

        ma = mask.Mask(
            array=np.array(
                [[False, False, True], [False, False, False], [True, False, False]]
            ),
            pixel_scale=1.0,
        )

        unmasked_sparse_grid_pixel_centres = np.array(
            [[0, 0], [0, 1], [0, 2], [0, 2], [0, 2], [1, 1]]
        )

        unmasked_sparse_to_sparse = mapping_util.unmasked_sparse_to_sparse_from_mask_and_pixel_centres(
            mask=ma,
            unmasked_sparse_grid_pixel_centres=unmasked_sparse_grid_pixel_centres,
            total_sparse_pixels=4,
        )

        assert (unmasked_sparse_to_sparse == np.array([0, 1, 2, 2, 2, 2])).all()

    def test__same_as_above__but_3x4_mask(self):

        ma = mask.Mask(
            array=np.array(
                [
                    [True, True, False, True],
                    [False, False, False, False],
                    [True, True, False, True],
                ]
            ),
            pixel_scale=1.0,
        )

        unmasked_sparse_grid_pixel_centres = np.array(
            [[0, 0], [0, 1], [0, 2], [0, 2], [0, 2], [1, 1], [2, 3], [0, 2]]
        )

        unmasked_sparse_to_sparse = mapping_util.unmasked_sparse_to_sparse_from_mask_and_pixel_centres(
            mask=ma,
            unmasked_sparse_grid_pixel_centres=unmasked_sparse_grid_pixel_centres,
            total_sparse_pixels=5,
        )

        assert (unmasked_sparse_to_sparse == np.array([0, 0, 0, 1, 2, 3, 4, 4])).all()

    def test__same_as_above__but_4x3_mask(self):

        ma = mask.Mask(
            array=np.array(
                [
                    [True, False, True],
                    [True, False, True],
                    [False, False, False],
                    [True, False, True],
                ]
            ),
            pixel_scale=1.0,
        )

        unmasked_sparse_grid_pixel_centres = np.array(
            [[0, 0], [0, 1], [0, 2], [0, 2], [0, 2], [1, 1], [2, 2], [3, 1]]
        )

        unmasked_sparse_to_sparse = mapping_util.unmasked_sparse_to_sparse_from_mask_and_pixel_centres(
            mask=ma,
            unmasked_sparse_grid_pixel_centres=unmasked_sparse_grid_pixel_centres,
            total_sparse_pixels=6,
        )

        assert (unmasked_sparse_to_sparse == np.array([0, 0, 1, 1, 1, 1, 2, 3])).all()


class TestRegularToSparse:
    def test__simple_cases_for_regular_to_unmasked_sparse_and__unmasked_sparse_to_sparse(
        self
    ):

        regular_to_unmasked_sparse = np.array([0, 1, 2, 3, 4])
        unmasked_sparse_to_sparse = np.array([0, 1, 2, 3, 4])
        regular_to_sparse = mapping_util.regular_to_sparse_from_sparse_mappings(
            regular_to_unmasked_sparse=regular_to_unmasked_sparse,
            unmasked_sparse_to_sparse=unmasked_sparse_to_sparse,
        )

        assert (regular_to_sparse == np.array([0, 1, 2, 3, 4])).all()

        regular_to_unmasked_sparse = np.array([0, 1, 2, 3, 4])
        unmasked_sparse_to_sparse = np.array([0, 1, 5, 7, 18])
        regular_to_sparse = mapping_util.regular_to_sparse_from_sparse_mappings(
            regular_to_unmasked_sparse=regular_to_unmasked_sparse,
            unmasked_sparse_to_sparse=unmasked_sparse_to_sparse,
        )

        assert (regular_to_sparse == np.array([0, 1, 5, 7, 18])).all()

        regular_to_unmasked_sparse = np.array([1, 1, 1, 1, 2])
        unmasked_sparse_to_sparse = np.array([0, 10, 15, 3, 4])
        regular_to_sparse = mapping_util.regular_to_sparse_from_sparse_mappings(
            regular_to_unmasked_sparse=regular_to_unmasked_sparse,
            unmasked_sparse_to_sparse=unmasked_sparse_to_sparse,
        )

        assert (regular_to_sparse == np.array([10, 10, 10, 10, 15])).all()

        regular_to_unmasked_sparse = np.array([5, 6, 7, 8, 9])
        unmasked_sparse_to_sparse = np.array([0, 1, 2, 3, 4, 19, 18, 17, 16, 15])
        regular_to_sparse = mapping_util.regular_to_sparse_from_sparse_mappings(
            regular_to_unmasked_sparse=regular_to_unmasked_sparse,
            unmasked_sparse_to_sparse=unmasked_sparse_to_sparse,
        )

        assert (regular_to_sparse == np.array([19, 18, 17, 16, 15])).all()


class TestSparseGridFromUnmaskedSparseGrid:
    def test__simple_unmasked_sparse_grid__full_grid_pix_grid_same_size__straightforward_mappings(
        self
    ):

        unmasked_sparse_grid = np.array(
            [[0.0, 0.0], [1.0, 1.0], [2.0, 2.0], [3.0, 3.0]]
        )
        sparse_to_unmasked_sparse = np.array([0, 1, 2, 3])
        pix_grid = mapping_util.sparse_grid_from_unmasked_sparse_grid(
            unmasked_sparse_grid=unmasked_sparse_grid,
            sparse_to_unmasked_sparse=sparse_to_unmasked_sparse,
        )

        assert (
            pix_grid == np.array([[0.0, 0.0], [1.0, 1.0], [2.0, 2.0], [3.0, 3.0]])
        ).all()

        unmasked_sparse_grid = np.array(
            [[0.0, 0.0], [4.0, 5.0], [2.0, 2.0], [8.0, 7.0]]
        )
        sparse_to_unmasked_sparse = np.array([0, 1, 2, 3])
        pix_grid = mapping_util.sparse_grid_from_unmasked_sparse_grid(
            unmasked_sparse_grid=unmasked_sparse_grid,
            sparse_to_unmasked_sparse=sparse_to_unmasked_sparse,
        )

        assert (
            pix_grid == np.array([[0.0, 0.0], [4.0, 5.0], [2.0, 2.0], [8.0, 7.0]])
        ).all()

        unmasked_sparse_grid = np.array(
            [[0.0, 0.0], [4.0, 5.0], [2.0, 2.0], [8.0, 7.0]]
        )
        sparse_to_unmasked_sparse = np.array([1, 0, 3, 2])
        pix_grid = mapping_util.sparse_grid_from_unmasked_sparse_grid(
            unmasked_sparse_grid=unmasked_sparse_grid,
            sparse_to_unmasked_sparse=sparse_to_unmasked_sparse,
        )

        assert (
            pix_grid == np.array([[4.0, 5.0], [0.0, 0.0], [8.0, 7.0], [2.0, 2.0]])
        ).all()

    def test__simple_unmasked_sparse_grid__full_grid_pix_bigger_than_pix__straightforward_mappings(
        self
    ):

        unmasked_sparse_grid = np.array(
            [[0.0, 0.0], [1.0, 1.0], [2.0, 2.0], [3.0, 3.0]]
        )
        sparse_to_unmasked_sparse = np.array([1, 2])
        pix_grid = mapping_util.sparse_grid_from_unmasked_sparse_grid(
            unmasked_sparse_grid=unmasked_sparse_grid,
            sparse_to_unmasked_sparse=sparse_to_unmasked_sparse,
        )

        assert (pix_grid == np.array([[1.0, 1.0], [2.0, 2.0]])).all()

        unmasked_sparse_grid = np.array(
            [[0.0, 0.0], [4.0, 5.0], [2.0, 2.0], [8.0, 7.0]]
        )
        sparse_to_unmasked_sparse = np.array([2, 2, 3])
        pix_grid = mapping_util.sparse_grid_from_unmasked_sparse_grid(
            unmasked_sparse_grid=unmasked_sparse_grid,
            sparse_to_unmasked_sparse=sparse_to_unmasked_sparse,
        )

        assert (pix_grid == np.array([[2.0, 2.0], [2.0, 2.0], [8.0, 7.0]])).all()

        unmasked_sparse_grid = np.array(
            [
                [0.0, 0.0],
                [4.0, 5.0],
                [2.0, 2.0],
                [8.0, 7.0],
                [11.0, 11.0],
                [-20.0, -15.0],
            ]
        )
        sparse_to_unmasked_sparse = np.array([1, 0, 5, 2])
        pix_grid = mapping_util.sparse_grid_from_unmasked_sparse_grid(
            unmasked_sparse_grid=unmasked_sparse_grid,
            sparse_to_unmasked_sparse=sparse_to_unmasked_sparse,
        )

        assert (
            pix_grid == np.array([[4.0, 5.0], [0.0, 0.0], [-20.0, -15.0], [2.0, 2.0]])
        ).all()



import autofit as af
import matplotlib

backend = af.conf.instance.visualize.get("figures", "backend", str)
matplotlib.use(backend)
from matplotlib import pyplot as plt

from autolens.data.plotters import data_plotters
from autolens.plotters import plotter_util


def plot_ccd_subplot(
    ccd_data,
    plot_origin=True,
    mask=None,
    extract_array_from_mask=False,
    zoom_around_mask=False,
    should_plot_border=False,
    positions=None,
    units="arcsec",
    kpc_per_arcsec=None,
    figsize=None,
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    titlesize=10,
    xlabelsize=10,
    ylabelsize=10,
    xyticksize=10,
    mask_pointsize=10,
    position_pointsize=30,
    grid_pointsize=1,
    output_path=None,
    output_filename="ccd_data",
    output_format="show",
):
    """Plot the ccd instrument as a sub-plot of all its quantites (e.g. the instrument, noise_map-map, PSF, Signal-to_noise-map, \
     etc).

    Set *autolens.instrument.array.plotters.array_plotters* for a description of all innput parameters not described below.

    Parameters
    -----------
    ccd_data : instrument.CCDData
        The ccd instrument, which includes the observed instrument, noise_map-map, PSF, signal-to-noise_map-map, etc.
    plot_origin : True
        If true, the origin of the instrument's coordinate system is plotted as a 'x'.
    image_plane_pix_grid : ndarray or instrument.array.grid_stacks.PixGrid
        If an adaptive pixelization whose pixels are formed by tracing pixels from the instrument, this plots those pixels \
        over the immage.
    ignore_config : bool
        If *False*, the config file general.ini is used to determine whether the subpot is plotted. If *True*, the \
        config file is ignored.
    """

    rows, columns, figsize_tool = plotter_util.get_subplot_rows_columns_figsize(
        number_subplots=6
    )

    if figsize is None:
        figsize = figsize_tool

    plt.figure(figsize=figsize)
    plt.subplot(rows, columns, 1)

    plot_image(
        ccd_data=ccd_data,
        plot_origin=plot_origin,
        mask=mask,
        extract_array_from_mask=extract_array_from_mask,
        zoom_around_mask=zoom_around_mask,
        should_plot_border=should_plot_border,
        positions=positions,
        as_subplot=True,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        mask_pointsize=mask_pointsize,
        position_pointsize=position_pointsize,
        grid_pointsize=grid_pointsize,
        output_path=output_path,
        output_format=output_format,
    )

    plt.subplot(rows, columns, 2)

    plot_noise_map(
        ccd_data=ccd_data,
        plot_origin=plot_origin,
        mask=mask,
        extract_array_from_mask=extract_array_from_mask,
        zoom_around_mask=zoom_around_mask,
        as_subplot=True,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        mask_pointsize=mask_pointsize,
        output_path=output_path,
        output_format=output_format,
    )

    plt.subplot(rows, columns, 3)

    plot_psf(
        ccd_data=ccd_data,
        as_subplot=True,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        output_path=output_path,
        output_format=output_format,
    )

    plt.subplot(rows, columns, 4)

    plot_signal_to_noise_map(
        ccd_data=ccd_data,
        plot_origin=plot_origin,
        mask=mask,
        extract_array_from_mask=extract_array_from_mask,
        zoom_around_mask=zoom_around_mask,
        as_subplot=True,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        mask_pointsize=mask_pointsize,
        output_path=output_path,
        output_format=output_format,
    )

    plt.subplot(rows, columns, 5)

    plot_absolute_signal_to_noise_map(
        ccd_data=ccd_data,
        plot_origin=plot_origin,
        mask=mask,
        extract_array_from_mask=extract_array_from_mask,
        zoom_around_mask=zoom_around_mask,
        as_subplot=True,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        mask_pointsize=mask_pointsize,
        output_path=output_path,
        output_format=output_format,
    )

    plt.subplot(rows, columns, 6)

    plot_potential_chi_squared_map(
        ccd_data=ccd_data,
        plot_origin=plot_origin,
        mask=mask,
        extract_array_from_mask=extract_array_from_mask,
        zoom_around_mask=zoom_around_mask,
        as_subplot=True,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        mask_pointsize=mask_pointsize,
        output_path=output_path,
        output_format=output_format,
    )

    plotter_util.output_subplot_array(
        output_path=output_path,
        output_filename=output_filename,
        output_format=output_format,
    )

    plt.close()


def plot_ccd_individual(
    ccd_data,
    plot_origin=True,
    mask=None,
    extract_array_from_mask=False,
    zoom_around_mask=False,
    positions=None,
    should_plot_image=False,
    should_plot_noise_map=False,
    should_plot_psf=False,
    should_plot_signal_to_noise_map=False,
    should_plot_absolute_signal_to_noise_map=False,
    should_plot_potential_chi_squared_map=False,
    units="arcsec",
    output_path=None,
    output_format="png",
):
    """Plot each attribute of the ccd instrument as individual figures one by one (e.g. the instrument, noise_map-map, PSF, \
     Signal-to_noise-map, etc).

    Set *autolens.instrument.array.plotters.array_plotters* for a description of all innput parameters not described below.

    Parameters
    -----------
    ccd_data : instrument.CCDData
        The ccd instrument, which includes the observed instrument, noise_map-map, PSF, signal-to-noise_map-map, etc.
    plot_origin : True
        If true, the origin of the instrument's coordinate system is plotted as a 'x'.
    """

    if should_plot_image:

        plot_image(
            ccd_data=ccd_data,
            plot_origin=plot_origin,
            mask=mask,
            extract_array_from_mask=extract_array_from_mask,
            zoom_around_mask=zoom_around_mask,
            positions=positions,
            units=units,
            output_path=output_path,
            output_format=output_format,
        )

    if should_plot_noise_map:

        plot_noise_map(
            ccd_data=ccd_data,
            plot_origin=plot_origin,
            mask=mask,
            extract_array_from_mask=extract_array_from_mask,
            zoom_around_mask=zoom_around_mask,
            units=units,
            output_path=output_path,
            output_format=output_format,
        )

    if should_plot_psf:

        plot_psf(
            ccd_data=ccd_data,
            plot_origin=plot_origin,
            output_path=output_path,
            output_format=output_format,
        )

    if should_plot_signal_to_noise_map:

        plot_signal_to_noise_map(
            ccd_data=ccd_data,
            plot_origin=plot_origin,
            mask=mask,
            extract_array_from_mask=extract_array_from_mask,
            zoom_around_mask=zoom_around_mask,
            units=units,
            output_path=output_path,
            output_format=output_format,
        )

    if should_plot_absolute_signal_to_noise_map:

        plot_absolute_signal_to_noise_map(
            ccd_data=ccd_data,
            plot_origin=plot_origin,
            mask=mask,
            extract_array_from_mask=extract_array_from_mask,
            zoom_around_mask=zoom_around_mask,
            units=units,
            output_path=output_path,
            output_format=output_format,
        )

    if should_plot_potential_chi_squared_map:

        plot_potential_chi_squared_map(
            ccd_data=ccd_data,
            plot_origin=plot_origin,
            mask=mask,
            extract_array_from_mask=extract_array_from_mask,
            zoom_around_mask=zoom_around_mask,
            units=units,
            output_path=output_path,
            output_format=output_format,
        )


def plot_image(
    ccd_data,
    plot_origin=True,
    grid=None,
    mask=None,
    extract_array_from_mask=False,
    zoom_around_mask=False,
    should_plot_border=False,
    positions=None,
    as_subplot=False,
    units="arcsec",
    kpc_per_arcsec=None,
    figsize=(7, 7),
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    title="CCD Image",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    mask_pointsize=10,
    position_pointsize=30,
    grid_pointsize=1,
    output_path=None,
    output_format="show",
    output_filename="ccd_image",
):
    """Plot the observed instrument of the ccd instrument.

    Set *autolens.instrument.array.plotters.array_plotters* for a description of all innput parameters not described below.

    Parameters
    -----------
    image : instrument.CCDData
        The ccd instrument, which includes the observed instrument, noise_map-map, PSF, signal-to-noise_map-map, etc.
    plot_origin : True
        If true, the origin of the instrument's coordinate system is plotted as a 'x'.
    image_plane_pix_grid : ndarray or instrument.array.grid_stacks.PixGrid
        If an adaptive pixelization whose pixels are formed by tracing pixels from the instrument, this plots those pixels \
        over the immage.
    """
    data_plotters.plot_image(
        image=ccd_data.image,
        plot_origin=plot_origin,
        grid=grid,
        mask=mask,
        extract_array_from_mask=extract_array_from_mask,
        zoom_around_mask=zoom_around_mask,
        should_plot_border=should_plot_border,
        positions=positions,
        as_subplot=as_subplot,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        title=title,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        mask_pointsize=mask_pointsize,
        position_pointsize=position_pointsize,
        grid_pointsize=grid_pointsize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )


def plot_noise_map(
    ccd_data,
    plot_origin=True,
    mask=None,
    extract_array_from_mask=False,
    zoom_around_mask=False,
    as_subplot=False,
    units="arcsec",
    kpc_per_arcsec=None,
    figsize=(7, 7),
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    title="CCD Noise-Map",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    mask_pointsize=10,
    output_path=None,
    output_format="show",
    output_filename="ccd_noise_map",
):
    """Plot the noise_map-map of the ccd instrument.

    Set *autolens.instrument.array.plotters.array_plotters* for a description of all innput parameters not described below.

    Parameters
    -----------
    image : instrument.CCDData
        The ccd instrument, which includes the observed instrument, noise_map-map, PSF, signal-to-noise_map-map, etc.
    plot_origin : True
        If true, the origin of the instrument's coordinate system is plotted as a 'x'.
    """

    data_plotters.plot_noise_map(
        noise_map=ccd_data.noise_map,
        plot_origin=plot_origin,
        mask=mask,
        extract_array_from_mask=extract_array_from_mask,
        zoom_around_mask=zoom_around_mask,
        as_subplot=as_subplot,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        title=title,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        mask_pointsize=mask_pointsize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )


def plot_psf(
    ccd_data,
    plot_origin=True,
    as_subplot=False,
    units="arcsec",
    kpc_per_arcsec=None,
    figsize=(7, 7),
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    title="CCD PSF",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    output_path=None,
    output_format="show",
    output_filename="ccd_psf",
):
    """Plot the PSF of the ccd instrument.

    Set *autolens.instrument.array.plotters.array_plotters* for a description of all innput parameters not described below.

    Parameters
    -----------
    image : instrument.CCDData
        The ccd instrument, which includes the observed instrument, noise_map-map, PSF, signal-to-noise_map-map, etc.
    plot_origin : True
        If true, the origin of the instrument's coordinate system is plotted as a 'x'.
    """

    data_plotters.plot_psf(
        psf=ccd_data.psf,
        plot_origin=plot_origin,
        as_subplot=as_subplot,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        title=title,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )


def plot_signal_to_noise_map(
    ccd_data,
    plot_origin=True,
    mask=None,
    extract_array_from_mask=False,
    zoom_around_mask=False,
    as_subplot=False,
    units="arcsec",
    kpc_per_arcsec=None,
    figsize=(7, 7),
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    title="CCD Signal-To-Noise-Map",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    mask_pointsize=10,
    output_path=None,
    output_format="show",
    output_filename="ccd_signal_to_noise_map",
):
    """Plot the signal-to-noise_map-map of the ccd instrument.

    Set *autolens.instrument.array.plotters.array_plotters* for a description of all innput parameters not described below.

    Parameters
    -----------
    image : instrument.CCDData
        The ccd instrument, which includes the observed image, noise_map-map, PSF, signal-to-noise_map-map, etc.
    plot_origin : True
        If true, the origin of the instrument's coordinate system is plotted as a 'x'.
    """

    data_plotters.plot_signal_to_noise_map(
        signal_to_noise_map=ccd_data.signal_to_noise_map,
        plot_origin=plot_origin,
        mask=mask,
        extract_array_from_mask=extract_array_from_mask,
        zoom_around_mask=zoom_around_mask,
        as_subplot=as_subplot,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        title=title,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        mask_pointsize=mask_pointsize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )


def plot_absolute_signal_to_noise_map(
    ccd_data,
    plot_origin=True,
    mask=None,
    extract_array_from_mask=False,
    zoom_around_mask=False,
    as_subplot=False,
    units="arcsec",
    kpc_per_arcsec=None,
    figsize=(7, 7),
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    title="CCD Absolute Signal-To-Noise-Map",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    mask_pointsize=10,
    output_path=None,
    output_format="show",
    output_filename="ccd_absolute_signal_to_noise_map",
):
    """Plot the signal-to-noise_map-map of the ccd instrument.

    Set *autolens.instrument.array.plotters.array_plotters* for a description of all innput parameters not described below.

    Parameters
    -----------
    image : instrument.CCDData
        The ccd instrument, which includes the observed image, noise_map-map, PSF, signal-to-noise_map-map, etc.
    plot_origin : True
        If true, the origin of the instrument's coordinate system is plotted as a 'x'.
    """

    data_plotters.plot_absolute_signal_to_noise_map(
        absolute_signal_to_noise_map=ccd_data.absolute_signal_to_noise_map,
        plot_origin=plot_origin,
        mask=mask,
        extract_array_from_mask=extract_array_from_mask,
        zoom_around_mask=zoom_around_mask,
        as_subplot=as_subplot,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        title=title,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        mask_pointsize=mask_pointsize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )


def plot_potential_chi_squared_map(
    ccd_data,
    plot_origin=True,
    mask=None,
    extract_array_from_mask=False,
    zoom_around_mask=False,
    as_subplot=False,
    units="arcsec",
    kpc_per_arcsec=None,
    figsize=(7, 7),
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    title="CCD Potential Chi-Squared Map",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    mask_pointsize=10,
    output_path=None,
    output_format="show",
    output_filename="ccd_potential_chi_squared_map",
):
    """Plot the signal-to-noise_map-map of the ccd instrument.

    Set *autolens.instrument.array.plotters.array_plotters* for a description of all innput parameters not described below.

    Parameters
    -----------
    image : instrument.CCDData
        The ccd instrument, which includes the observed image, noise_map-map, PSF, signal-to-noise_map-map, etc.
    plot_origin : True
        If true, the origin of the instrument's coordinate system is plotted as a 'x'.
    """

    data_plotters.plot_potential_chi_squared_map(
        potential_chi_squared_map=ccd_data.potential_chi_squared_map,
        plot_origin=plot_origin,
        mask=mask,
        extract_array_from_mask=extract_array_from_mask,
        zoom_around_mask=zoom_around_mask,
        as_subplot=as_subplot,
        units=units,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        title=title,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        mask_pointsize=mask_pointsize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )


def get_origin(array, plot_origin):
    """Get the (y,x) origin of the ccd instrument if it going to be plotted.

    Parameters
    -----------
    array : instrument.array.scaled_array.ScaledArray
        The array from which the origin is extracted.
    plot_origin : True
        If true, the origin of the instrument's coordinate system is returned.
    """
    if plot_origin:
        return array.origin
    else:
        return None



from autolens import decorator_util
import numpy as np


@decorator_util.jit()
def data_vector_from_blurred_mapping_matrix_and_data(
    blurred_mapping_matrix, image_1d, noise_map_1d
):
    """Compute the hyper_galaxy vector *D* from a blurred mapping matrix *f* and the 1D image *d* and 1D noise-map *\sigma* \
    (see Warren & Dye 2003).
    
    Parameters
    -----------
    blurred_mapping_matrix : ndarray
        The matrix representing the blurred mappings between sub-grid pixels and pixelization pixels.
    image_1d : ndarray
        Flattened 1D array of the observed image the inversion is fitting.
    noise_map_1d : ndarray
        Flattened 1D array of the noise-map used by the inversion during the fit.
    """

    mapping_shape = blurred_mapping_matrix.shape

    data_vector = np.zeros(mapping_shape[1])

    for image_index in range(mapping_shape[0]):
        for pix_index in range(mapping_shape[1]):
            data_vector[pix_index] += (
                image_1d[image_index]
                * blurred_mapping_matrix[image_index, pix_index]
                / (noise_map_1d[image_index] ** 2.0)
            )

    return data_vector


def curvature_matrix_from_blurred_mapping_matrix(blurred_mapping_matrix, noise_map_1d):
    """Compute the curvature matrix *F* from a blurred mapping matrix *f* and the 1D noise-map *\sigma* \
     (see Warren & Dye 2003).

    Parameters
    -----------
    blurred_mapping_matrix : ndarray
        The matrix representing the blurred mappings between sub-grid pixels and pixelization pixels.
    noise_map_1d : ndarray
        Flattened 1D array of the noise-map used by the inversion during the fit.
    """

    flist = np.zeros(blurred_mapping_matrix.shape[0])
    iflist = np.zeros(blurred_mapping_matrix.shape[0], dtype="int")
    return curvature_matrix_from_blurred_mapping_matrix_jit(
        blurred_mapping_matrix, noise_map_1d, flist, iflist
    )


@decorator_util.jit()
def curvature_matrix_from_blurred_mapping_matrix_jit(
    blurred_mapping_matrix, noise_map_1d, flist, iflist
):
    """Compute the curvature matrix *F* from a blurred mapping matrix *f* and the 1D noise-map *\sigma* \
    (see Warren & Dye 2003).

    Parameters
    -----------
    blurred_mapping_matrix : ndarray
        The matrix representing the blurred mappings between sub-grid pixels and pixelization pixels.
    noise_map_1d : ndarray
        Flattened 1D array of the noise-map used by the inversion during the fit.
    flist : ndarray
        NumPy array of floats used to store mappings for efficienctly calculation.
    iflist : ndarray
        NumPy array of integers used to store mappings for efficienctly calculation.
    """
    curvature_matrix = np.zeros(
        (blurred_mapping_matrix.shape[1], blurred_mapping_matrix.shape[1])
    )

    for image_index in range(blurred_mapping_matrix.shape[0]):
        index = 0
        for pixel_index in range(blurred_mapping_matrix.shape[1]):
            if blurred_mapping_matrix[image_index, pixel_index] > 0.0:
                flist[index] = (
                    blurred_mapping_matrix[image_index, pixel_index]
                    / noise_map_1d[image_index]
                )
                iflist[index] = pixel_index
                index += 1

        if index > 0:
            for i1 in range(index):
                for j1 in range(index):
                    ix = iflist[i1]
                    iy = iflist[j1]
                    curvature_matrix[ix, iy] += flist[i1] * flist[j1]

    for i in range(blurred_mapping_matrix.shape[1]):
        for j in range(blurred_mapping_matrix.shape[1]):
            curvature_matrix[i, j] = curvature_matrix[j, i]

    return curvature_matrix


@decorator_util.jit()
def reconstructed_data_vector_from_blurred_mapping_matrix_and_solution_vector(
    blurred_mapping_matrix, solution_vector
):
    """ Compute the reconstructed hyper_galaxy vector from the blurrred mapping matrix *f* and solution vector *S*.

    Parameters
    -----------
    blurred_mapping_matrix : ndarray
        The matrix representing the blurred mappings between sub-grid pixels and pixelization pixels.

    """
    reconstructed_data_vector = np.zeros(blurred_mapping_matrix.shape[0])
    for i in range(blurred_mapping_matrix.shape[0]):
        for j in range(solution_vector.shape[0]):
            reconstructed_data_vector[i] += (
                solution_vector[j] * blurred_mapping_matrix[i, j]
            )

    return reconstructed_data_vector


# @decorator_util.jit()
def pixelization_residual_map_from_pixelization_values_and_reconstructed_data_1d(
    pixelization_values, reconstructed_data_1d, sub_to_regular, pixelization_to_sub_all
):

    pixelization_residuals = np.zeros(shape=len(pixelization_to_sub_all))

    reconstructed_data_1d = reconstructed_data_1d

    for pixelization_index, sub_pixels in enumerate(pixelization_to_sub_all):
        for sub_index in sub_pixels:
            regular_index = sub_to_regular[sub_index]
            residual = (
                reconstructed_data_1d[regular_index]
                - pixelization_values[pixelization_index]
            )
            pixelization_residuals[pixelization_index] += residual

    return pixelization_residuals


# @decorator_util.jit()
def pixelization_normalized_residual_map_from_pixelization_values_and_reconstructed_data_1d(
    pixelization_values,
    reconstructed_data_1d,
    noise_map_1d,
    sub_to_regular,
    pixelization_to_sub_all,
):

    pixelization_normalized_residuals = np.zeros(shape=len(pixelization_to_sub_all))

    reconstructed_data_1d = reconstructed_data_1d

    for pixelization_index, sub_pixels in enumerate(pixelization_to_sub_all):
        for sub_index in sub_pixels:
            regular_index = sub_to_regular[sub_index]
            residual = (
                reconstructed_data_1d[regular_index]
                - pixelization_values[pixelization_index]
            )
            pixelization_normalized_residuals[pixelization_index] += (
                residual / noise_map_1d[regular_index]
            )

    return pixelization_normalized_residuals


# @decorator_util.jit()
def pixelization_chi_squared_map_from_pixelization_values_and_reconstructed_data_1d(
    pixelization_values,
    reconstructed_data_1d,
    noise_map_1d,
    sub_to_regular,
    pixelization_to_sub_all,
):

    pixelization_chi_squareds = np.zeros(shape=len(pixelization_to_sub_all))

    reconstructed_data_1d = reconstructed_data_1d

    for pixelization_index, sub_pixels in enumerate(pixelization_to_sub_all):
        for sub_index in sub_pixels:
            regular_index = sub_to_regular[sub_index]
            residual = (
                reconstructed_data_1d[regular_index]
                - pixelization_values[pixelization_index]
            )
            pixelization_chi_squareds[pixelization_index] += (
                residual / noise_map_1d[regular_index]
            ) ** 2.0

    return pixelization_chi_squareds



import getdist
import re
import os
import math
import numpy as np
from astropy import cosmology as cosmo
import matplotlib.pyplot as plt

from autolens.data.array import grids
from autolens.model.galaxy import galaxy
from autolens.model.profiles import light_profiles
from autolens.model.profiles import mass_profiles
from autolens.lens import ray_tracing


class SLACS(object):
    def __init__(self):

        cosmological_model = cosmo.LambdaCDM(H0=70, Om0=0.3, Ode0=0.7)
        cosmological_model = cosmo.Planck15

        self.redshift = 0.169
        self.source_redshift = 0.451
        self.critical_density_arcsec = 40246450650.700836
        self.einstein_radius = 1.91716
        self.source_light_min = 0.1
        self.source_light_max = 2.5

        self.arcsec_per_kpc = cosmological_model.arcsec_per_kpc(self.redshift).value
        self.kpc_per_arcsec = 1.0 / self.arcsec_per_kpc

        # stop

    def setup_lensing_plane_sie(self, values):

        total_mass = mass_profiles.EllipticalIsothermal(
            centre=(values[0], values[1]),
            einstein_radius=values[2],
            axis_ratio=values[3],
            phi=values[4],
        )

        self.lens_galaxy = galaxy.Galaxy(redshift=self.redshift, sie=total_mass)
        self.source_galaxy = galaxy.Galaxy(redshift=self.source_redshift)

    def setup_lensing_plane_ltm2(self, values, center_skip, ltm_skip):

        sersic_bulge_light = light_profiles.EllipticalSersic(
            centre=(values[0], values[1]),
            axis_ratio=values[5],
            phi=values[6],
            intensity=values[2],
            effective_radius=values[3],
            sersic_index=values[4],
        )

        exponential_halo_light = light_profiles.EllipticalExponential(
            axis_ratio=values[9 + center_skip],
            phi=values[10 + center_skip],
            intensity=values[7 + center_skip],
            effective_radius=values[8 + center_skip],
        )

        sersic_bulge = mass_profiles.EllipticalSersic(
            centre=(values[0], values[1]),
            axis_ratio=values[5],
            phi=values[6],
            intensity=values[2],
            effective_radius=values[3],
            sersic_index=values[4],
            mass_to_light_ratio=values[12 + center_skip],
        )

        exponential_halo = mass_profiles.EllipticalExponential(
            axis_ratio=values[9 + center_skip],
            phi=values[10 + center_skip],
            intensity=values[7 + center_skip],
            effective_radius=values[8 + center_skip],
            mass_to_light_ratio=values[12 + ltm_skip + center_skip],
        )

        dark_matter_halo = mass_profiles.SphericalNFW(
            kappa_s=values[11 + center_skip], scale_radius=30.0 * self.arcsec_per_kpc
        )

        self.lens_galaxy = galaxy.Galaxy(
            redshift=self.redshift,
            light1=sersic_bulge_light,
            light2=exponential_halo_light,
            bulge=sersic_bulge,
            envelope=exponential_halo,
            halo=dark_matter_halo,
        )

        self.source_galaxy = galaxy.Galaxy(redshift=self.source_redshift)

    def setup_lensing_plane_ltm2r_bulge(self, values, center_skip, ltm_skip):

        sersic_bulge_light = light_profiles.EllipticalSersic(
            centre=(values[0], values[1]),
            axis_ratio=values[5],
            phi=values[6],
            intensity=values[2],
            effective_radius=values[3],
            sersic_index=values[4],
        )

        exponential_halo_light = light_profiles.EllipticalExponential(
            axis_ratio=values[9 + center_skip],
            phi=values[10 + center_skip],
            intensity=values[7 + center_skip],
            effective_radius=values[8 + center_skip],
        )

        sersic_bulge = mass_profiles.EllipticalSersicRadialGradient(
            centre=(values[0], values[1]),
            axis_ratio=values[5],
            phi=values[6],
            intensity=values[2],
            effective_radius=values[3],
            sersic_index=values[4],
            mass_to_light_ratio=values[12 + center_skip],
            mass_to_light_gradient=values[13 + ltm_skip + center_skip],
        )

        exponential_halo = mass_profiles.EllipticalSersic(
            axis_ratio=values[9 + center_skip],
            phi=values[10 + center_skip],
            intensity=values[7 + center_skip],
            effective_radius=values[8 + center_skip],
            sersic_index=1.0,
            mass_to_light_ratio=values[12 + ltm_skip + center_skip],
        )

        dark_matter_halo = mass_profiles.SphericalNFW(
            kappa_s=values[11 + center_skip], scale_radius=30.0 * self.arcsec_per_kpc
        )

        self.lens_galaxy = galaxy.Galaxy(
            redshift=self.redshift,
            light1=sersic_bulge_light,
            light2=exponential_halo_light,
            bulge=sersic_bulge,
            envelope=exponential_halo,
            halo=dark_matter_halo,
        )

        self.source_galaxy = galaxy.Galaxy(redshift=self.source_redshift)

    def setup_lensing_plane_ltm2r_envelope(self, values, center_skip, ltm_skip):

        sersic_bulge_light = light_profiles.EllipticalSersic(
            centre=(values[0], values[1]),
            axis_ratio=values[5],
            phi=values[6],
            intensity=values[2],
            effective_radius=values[3],
            sersic_index=values[4],
        )

        exponential_halo_light = light_profiles.EllipticalExponential(
            axis_ratio=values[9 + center_skip],
            phi=values[10 + center_skip],
            intensity=values[7 + center_skip],
            effective_radius=values[8 + center_skip],
        )

        sersic_bulge = mass_profiles.EllipticalSersic(
            centre=(values[0], values[1]),
            axis_ratio=values[5],
            phi=values[6],
            intensity=values[2],
            effective_radius=values[3],
            sersic_index=values[4],
            mass_to_light_ratio=values[12 + center_skip],
        )

        exponential_halo = mass_profiles.EllipticalSersicRadialGradient(
            axis_ratio=values[9 + center_skip],
            phi=values[10 + center_skip],
            intensity=values[7 + center_skip],
            effective_radius=values[8 + center_skip],
            sersic_index=1.0,
            mass_to_light_ratio=values[12 + ltm_skip + center_skip],
            mass_to_light_gradient=values[13 + ltm_skip + center_skip],
        )

        dark_matter_halo = mass_profiles.SphericalNFW(
            kappa_s=values[11 + center_skip], scale_radius=30.0 * self.arcsec_per_kpc
        )

        self.lens_galaxy = galaxy.Galaxy(
            redshift=self.redshift,
            light1=sersic_bulge_light,
            light2=exponential_halo_light,
            bulge=sersic_bulge,
            envelope=exponential_halo,
            halo=dark_matter_halo,
        )

        self.source_galaxy = galaxy.Galaxy(redshift=self.source_redshift)

    def density_vs_radii_sie(self, values, radius_kpc, number_bins=60):

        self.setup_lensing_plane_sie(values)
        radii = list(
            np.linspace(5e-3, radius_kpc * self.arcsec_per_kpc, number_bins + 1)
        )

        self.sie_density_plot = []
        self.sie_density_upper_plot = []
        self.sie_density_lower_plot = []

        self.radii_plot = []

        for r in range(number_bins):

            annuli_area = math.pi * radii[r + 1] ** 2 - math.pi * radii[r] ** 2

            self.setup_lensing_plane_sie(values)

            sie_density = (
                self.critical_density_arcsec
                * (
                    self.lens_galaxy.sie.mass_within_circle_in_units(radii[r + 1])
                    - self.lens_galaxy.sie.mass_within_circle_in_units(radii[r])
                )
                / annuli_area
            )

            self.sie_density_plot.append(sie_density)

            self.radii_plot.append(
                ((radii[r + 1] + radii[r]) / 2.0) * self.kpc_per_arcsec
            )

    def density_vs_radii_ltm2(
        self, values, radius_kpc, center_skip, ltm_skip, gradient=None, number_bins=60
    ):

        if gradient is None:
            self.setup_lensing_plane_ltm2(values, center_skip, ltm_skip)
        elif gradient is "bulge":
            self.setup_lensing_plane_ltm2r_bulge(values, center_skip, ltm_skip)
        elif gradient is "envelope":
            self.setup_lensing_plane_ltm2r_envelope(values, center_skip, ltm_skip)

        radii = list(
            np.linspace(5e-3, radius_kpc * self.arcsec_per_kpc, number_bins + 1)
        )

        self.bulge_density_plot = []
        self.bulge_density_upper_plot = []
        self.bulge_density_lower_plot = []

        self.envelope_density_plot = []
        self.envelope_density_upper_plot = []
        self.envelope_density_lower_plot = []

        self.dark_density_plot = []
        self.dark_density_upper_plot = []
        self.dark_density_lower_plot = []

        self.radii_plot = []

        for r in range(number_bins):

            annuli_area = math.pi * radii[r + 1] ** 2 - math.pi * radii[r] ** 2

            if gradient is None:
                self.setup_lensing_plane_ltm2(values, center_skip, ltm_skip)
            elif gradient is "bulge":
                self.setup_lensing_plane_ltm2r_bulge(values, center_skip, ltm_skip)
            elif gradient is "envelope":
                self.setup_lensing_plane_ltm2r_envelope(values, center_skip, ltm_skip)

            bulge_density = (
                self.critical_density_arcsec
                * (
                    self.lens_galaxy.bulge.mass_within_circle_in_units(radii[r + 1])
                    - self.lens_galaxy.bulge.mass_within_circle_in_units(radii[r])
                )
                / annuli_area
            )

            envelope_density = (
                self.critical_density_arcsec
                * (
                    self.lens_galaxy.envelope.mass_within_circle_in_units(radii[r + 1])
                    - self.lens_galaxy.envelope.mass_within_circle_in_units(radii[r])
                )
                / annuli_area
            )

            halo_density = (
                self.critical_density_arcsec
                * (
                    self.lens_galaxy.halo.mass_within_circle_in_units(radii[r + 1])
                    - self.lens_galaxy.halo.mass_within_circle_in_units(radii[r])
                )
                / annuli_area
            )

            self.bulge_density_plot.append(bulge_density)
            self.envelope_density_plot.append(envelope_density)
            self.dark_density_plot.append(halo_density)

            self.radii_plot.append(
                ((radii[r + 1] + radii[r]) / 2.0) * self.kpc_per_arcsec
            )

    def plot_density(
        self,
        image_name="",
        file_name="file",
        title="",
        labels=None,
        xaxis_is_physical=True,
        yaxis_is_physical=True,
    ):

        plt.figure(figsize=(10, 10))
        plt.title(title + image_name, size=18)

        if xaxis_is_physical:
            plt.xlabel("Distance From Galaxy Center (kpc)", size=20)
        else:
            plt.xlabel('Distance From Galaxy Center (")', size=20)

        if yaxis_is_physical:
            plt.ylabel(
                r"Surface Mass Density $\Sigma$ ($M_{\odot}$ / kpc $^2$)", size=20
            )
        else:
            pass

        self.decomposed_density_plot = list(
            map(
                lambda bulge, envelope, dark: bulge + envelope + dark,
                self.bulge_density_plot,
                self.envelope_density_plot,
                self.dark_density_plot,
            )
        )

        plt.semilogy(self.radii_plot, self.sie_density_plot, color="g", label="SIE")
        plt.semilogy(
            self.radii_plot,
            self.decomposed_density_plot,
            color="c",
            label="Decomposed Total",
        )
        plt.semilogy(
            self.radii_plot,
            self.bulge_density_plot,
            color="r",
            linestyle="--",
            label="Sersic",
        )
        plt.semilogy(
            self.radii_plot,
            self.envelope_density_plot,
            color="b",
            linestyle="--",
            label="Exponential",
        )
        plt.semilogy(
            self.radii_plot,
            self.dark_density_plot,
            color="k",
            linestyle="--",
            label="Dark Matter",
        )
        #
        # plt.semilogy(self.radii_plot, self.bulge_density_upper_plot, color='r', linestyle='--')
        # plt.semilogy(self.radii_plot, self.halo_density_upper_plot, color='g', linestyle='--')
        # plt.semilogy(self.radii_plot, self.dark_density_upper_plot, color='k', linestyle='--')
        #
        # plt.semilogy(self.radii_plot, self.bulge_density_lower_plot, color='r', linestyle='--')
        # plt.semilogy(self.radii_plot, self.halo_density_lower_plot, color='g', linestyle='--')
        # plt.semilogy(self.radii_plot, self.dark_density_lower_plot, color='k', linestyle='--')

        # plt.loglog(radii_plot, density_0, color='r', label='Sersic Bulge')
        # plt.loglog(radii_plot, density_1, color='g', label='EllipticalExponentialMass Halo')
        # plt.loglog(radii_plot, density_2, color='k', label='Dark Matter Halo')

        # plt.semilogy(radii_plot, density_lower[:, 0], color='r', linestyle='--')
        # plt.semilogy(radii_plot, density_upper[:, 0], color='r', linestyle='--')
        # plt.semilogy(radii_plot, density_lower[:, 1], color='g', linestyle='--')
        # plt.semilogy(radii_plot, density_upper[:, 1], color='g', linestyle='--')
        # plt.semilogy(radii_plot, density_lower[:, 2], color='k', linestyle='--')
        # plt.semilogy(radii_plot, density_upper[:, 2], color='k', linestyle='--')

        plt.axvline(x=self.einstein_radius * self.kpc_per_arcsec, linestyle="--")
        plt.axvline(x=self.source_light_min * self.kpc_per_arcsec, linestyle="-")
        plt.axvline(x=self.source_light_max * self.kpc_per_arcsec, linestyle="-")

        plt.tick_params(labelsize=16)
        plt.legend(fontsize=11.5)

        import os

        path = "{}/".format(os.path.dirname(os.path.realpath(__file__)))
        plt.savefig(path + "plots/" + file_name, bbox_inches="tight")
        plt.show()

    def masses_of_all_samples_sie(self, radius_kpc, values):

        self.setup_lensing_plane_sie(values=values)

        radius_arc = radius_kpc * self.arcsec_per_kpc
        sie_mass = (
            self.critical_density_arcsec
            * self.lens_galaxy.sie.mass_within_circle_in_units(radius=radius_arc)
        )

        return sie_mass

    def masses_of_all_samples_ltm2(self, radius_kpc, values, center_skip, ltm_skip):

        self.setup_lensing_plane_ltm2(
            values=values, center_skip=center_skip, ltm_skip=ltm_skip
        )

        radius_arc = radius_kpc * self.arcsec_per_kpc
        bulge_mass = (
            self.critical_density_arcsec
            * self.lens_galaxy.bulge.mass_within_circle_in_units(radius=radius_arc)
        )
        envelope_mass = (
            self.critical_density_arcsec
            * self.lens_galaxy.envelope.mass_within_circle_in_units(radius=radius_arc)
        )
        halo_mass = (
            self.critical_density_arcsec
            * self.lens_galaxy.halo.mass_within_circle_in_units(radius=radius_arc)
        )

        total_mass = bulge_mass + envelope_mass + halo_mass
        stellar_mass = bulge_mass + envelope_mass
        dark_mass = halo_mass

        return total_mass, stellar_mass, bulge_mass, envelope_mass, dark_mass

    def masses_of_all_samples_ltm2r_bulge(
        self, radius_kpc, values, center_skip, ltm_skip
    ):

        self.setup_lensing_plane_ltm2r_bulge(
            values=values, center_skip=center_skip, ltm_skip=ltm_skip
        )

        radius_arc = radius_kpc * self.arcsec_per_kpc
        bulge_mass = (
            self.critical_density_arcsec
            * self.lens_galaxy.bulge.mass_within_circle_in_units(radius=radius_arc)
        )
        envelope_mass = (
            self.critical_density_arcsec
            * self.lens_galaxy.envelope.mass_within_circle_in_units(radius=radius_arc)
        )
        halo_mass = (
            self.critical_density_arcsec
            * self.lens_galaxy.halo.mass_within_circle_in_units(radius=radius_arc)
        )

        total_mass = bulge_mass + envelope_mass + halo_mass
        stellar_mass = bulge_mass + envelope_mass
        dark_mass = halo_mass

        return total_mass, stellar_mass, bulge_mass, envelope_mass, dark_mass

    def masses_of_all_samples_ltm2r_envelope(
        self, radius_kpc, values, center_skip, ltm_skip
    ):

        self.setup_lensing_plane_ltm2r_envelope(
            values=values, center_skip=center_skip, ltm_skip=ltm_skip
        )

        radius_arc = radius_kpc * self.arcsec_per_kpc
        bulge_mass = (
            self.critical_density_arcsec
            * self.lens_galaxy.bulge.mass_within_circle_in_units(radius=radius_arc)
        )
        envelope_mass = (
            self.critical_density_arcsec
            * self.lens_galaxy.envelope.mass_within_circle_in_units(radius=radius_arc)
        )
        halo_mass = (
            self.critical_density_arcsec
            * self.lens_galaxy.halo.mass_within_circle_in_units(radius=radius_arc)
        )

        total_mass = bulge_mass + envelope_mass + halo_mass
        stellar_mass = bulge_mass + envelope_mass
        dark_mass = halo_mass

        return total_mass, stellar_mass, bulge_mass, envelope_mass, dark_mass



from autolens.data.array import mask as msk

import numpy as np


class MockMask(msk.Mask):
    def __new__(cls, array, pixel_scale=1.0, *args, **kwargs):

        obj = np.array(array, dtype="bool").view(cls)
        obj.pixel_scale = pixel_scale
        obj.origin = (0.0, 0.0)

        return obj

    def __init__(self, array, pixel_scale=1.0):
        pass


class MockMask1D(np.ndarray):
    def __new__(cls, shape, pixel_scale=1.0, *args, **kwargs):

        array = np.full(fill_value=False, shape=shape)

        obj = np.array(array, dtype="bool").view(cls)
        obj.pixel_scale = pixel_scale
        obj.origin = (0.0, 0.0)

        return obj
